// %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
// %%% Various Shared Utility Routines %%%%%%%
// %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

// These utilities are common to multiple projects, and have found use in a number of
// common scenarios.  For this reason, they've been given their own package.
// 
// Installing:
//   npm install @pentestusa/custom_utils__nodejs
// 
// 
/// Usage: 
//   Due to these routines being used in the core audit cluster application package, which utilizes its own
//   dependency tree, they are instantiated differently than most node modules.  However, once instantiated they're 
//   usable the same way most other node modules are.
// 

/*
	var custom_utils = await require("custom_utils__nodejs/custom_utils.js")({});
*/


// this object will be exported from this module
var exported = {};

// set the app config as the module parameters (should always be the case)
var app_config = null;


// exporter function
module.exports = async (module_params = null) => {

// %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
// %%% Global Configs %%%%%%%%%%%%%%%%%%%%%%%%%
// %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

// set application configuration
app_config = module_params;


// %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
// %%% Third Party Dependencies %%%%%%%%%%%%%%
// %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

// standard dependencies
var decycle     = require('json-decycle').decycle;
var ip          = require("ip");
var ip_cidr     = require('ip-cidr');
var colors      = require("colors");
var validator   = require("validator");
var fs          = require("fs");
var StackTrace  = require("stacktrace-js");
var uuid_random = require("uuid-random");
var nested_keys = require('nested-keys');
var parsePng    = require('parse-png');
var netstat     = require('node-netstat');

// since this application uses worker threads, we need this 
// to determine if we're in a thread or not.
const {isMainThread} = require('worker_threads');



// %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
// %%% Module Internal Globals %%%%%%%
// %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

// db to use for logging
var logger_db = null;

// current run information
var run_info = 
{
    pid:      null,
    type:     null,
    run_id:   null,
    start_ts: null,
    log_counters:
    {
        log:  0,
        err:  0,
        warn: 0
    }
};



// %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
// %%% Character Checks %%%%%%%%%%%%%%
// %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

// simply checks if a buffer is all ascii characters
var isAsciiPrintable = function(str)
{
    var ascii_check_regexp = new RegExp('^[\x09\x0a\x0d\x20-\x7e]+$', "g");
    if(ascii_check_regexp.test(str) === true)
        return true;
    
    return false;
}

// check if a string contains alphanumeric or underscore
var isAlphaNumericUnderscoreSpace = function(str)
{
    var ascii_check_regexp = new RegExp('^[A-Za-z0-9_ ]+$', "g");
    if(ascii_check_regexp.test(str) === true)
        return true;
    
    return false;
}

// check if a string contains alphanumeric or underscore
var isAlphaNumericUnderscore = function(str)
{
    var ascii_check_regexp = new RegExp('^[A-Za-z0-9_]+$', "g");
    if(ascii_check_regexp.test(str) === true)
        return true;
    
    return false;
}

// check if a string contains alphanumeric values
var isAlphaNumeric = function(str)
{
    var ascii_check_regexp = new RegExp('^[A-Za-z0-9]+$', "g");
    if(ascii_check_regexp.test(str) === true)
        return true;
    
    return false;
}

// Check if a string contains alphabet characters only.
var isAlphabet = function(str)
{
    var ascii_check_regexp = new RegExp('^[A-Za-z]+$', "g");
    if(ascii_check_regexp.test(str) === true)
        return true;
    
    return false;
}

// export checks
exported.isAsciiPrintable              = isAsciiPrintable;
exported.isAlphaNumericUnderscoreSpace = isAlphaNumericUnderscoreSpace;
exported.isAlphaNumericUnderscore      = isAlphaNumericUnderscore;
exported.isAlphaNumeric                = isAlphaNumeric;
exported.isAlphabet                    = isAlphabet;


// %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
// %%% Chop/Chomp String %%%%%%%%%%%%%
// %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

// simply remove the last character in a string
var chop = function(str)
{
    return str.substr(0, str.length - 1);
}

// left trim
var ltrim = function(str)
{
    return str.substring(1);
}

// export utilities
exported.chop  = chop;
exported.rtrim = chop;
exported.lchop = ltrim;
exported.ltrim = ltrim;


// %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
// %%% Check if Async Function %%%%%%%%%%%%%%%%%%%%%%%
// %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

var isAsyncFunction = function(func)
{

    // check if a function is async
    var is_async_function = false;
    try
    {
        if(func.constructor.name == 'AsyncFunction')
            is_async_function = true;
    } catch(err){}

    // return true or false
    return is_async_function;

}

// export the test function
exported.isAsyncFunction = isAsyncFunction;


// %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
// %%% Class Information %%%%%%%%%%%%%%%%%%%%%%%%%%%%%
// %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

// get class methods
var getClassMethods = function(class_instance)
{

    // check class properties
    if(isEmpty(class_instance) === true)
        return null;
    if(typeof class_instance !== "object")
        return null;

    // get class methods 
    class_methods = Object.getOwnPropertyNames(class_instance.__proto__).filter(function(p)
    {
        return typeof class_instance[p] === "function";
    });

    // return the class methods
    return class_methods;

}

// export utilities
exported.getClassMethods = getClassMethods;


// %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
// %%% Simple Validators %%%%%%%%%%%%%%%%%%%%%%%%%%%%%
// %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

// check if value is a uuid
var isValidUUID = function(uuid_val)
{
   

    // ensure value isn't empty
    if(isEmpty(uuid_val) === true)
        return false;

    // check to ensure guid is valid double length guid
    if(/^[0-9a-f]{8}-[0-9a-f]{4}-[0-9a-f]{4}-[0-9a-f]{4}-[0-9a-f]{12}$/i.test(uuid_val) !== true)
    {
        return false;
    }

    // return indicating success
    return true;

}

// check for valid double length uuid
var isValidDoubleLengthUUID = function(double_length_uuid_val)
{

    // ensure value isn't empty
    if(isEmpty(double_length_uuid_val) === true)
        return false;

    // check to ensure guid is valid double length guid
    if(/^[0-9a-f]{8}-[0-9a-f]{4}-[0-9a-f]{4}-[0-9a-f]{4}-[0-9a-f]{12}[0-9a-f]{8}-[0-9a-f]{4}-[0-9a-f]{4}-[0-9a-f]{4}-[0-9a-f]{12}$/i.test(double_length_uuid_val) !== true)
    {
        return false;
    }

    // return indicating success
    return true;

}

// export utilities
exported.isValidUUID             = isValidUUID;
exported.isValidDoubleLengthUUID = isValidDoubleLengthUUID;


// %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
// %%% Array Sorting Utilities %%%%%%%%%%%%%%%%%%%%%%%
// %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

// sort the ipv4 array
var sortIPv4Array = function(ipv4_string_array)
{

    // ensure input is an array
    if(Array.isArray(ipv4_string_array) === false)
        return null;

    
    // sorted results
    var sorted_results = [];

    // iterate through data and add results
    for(var idx = 0; idx < ipv4_string_array.length; idx++)
    {
        try
        {
            sorted_results.push(ip.toLong(ipv4_string_array[idx]));
        } catch(err){}
    }

    // sort the results by integer
    sorted_results.sort(function(a, b){ return a - b;});

    // iterate through results
    for(var idx = 0; idx < sorted_results.length; idx++)
    {
        sorted_results[idx] = ip.fromLong(sorted_results[idx]);
    }

    // return sorted results
    return sorted_results;

}


// sort integer string array (parses to in, then sorts by string)
var sortIntegerStringArray = function(int_string_array)
{

    // ensure input is an array
    if(Array.isArray(int_string_array) === false)
        return null;

    
    // sorted results
    var sorted_results = [];

    // iterate through data and add results
    for(var idx = 0; idx < int_string_array.length; idx++)
    {
        try
        {
            sorted_results.push(parseInt(int_string_array[idx]));
        } catch(err){}
    }

    // sort the results by integer
    sorted_results.sort(function(a, b){ return a - b;});
    
    // iterate through results
    for(var idx = 0; idx < sorted_results.length; idx++)
    {
        sorted_results[idx] = sorted_results[idx].toString();
    }

    // return sorted results
    return sorted_results;

}

// export sorting utilities 
exported.sortIPv4Array          = sortIPv4Array;
exported.sortIntegerStringArray = sortIntegerStringArray;


// %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
// %%% Time Utils %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
// %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

// Sleep asynchronously, resolving the returned promise only after a timeout executes.
var asyncSleep = function(seconds)
{
    return new Promise(function(resolve, reject)
    {
        setTimeout
        (
            function()
            {
                resolve();
            }, 
            seconds * 1000
        );
    });
}

// Sleep asynchronously (in miliseconds), resolving the returned promise only after a timeout executes.
var asyncSleepMiliseconds = function(miliseconds)
{
    return new Promise(function(resolve, reject)
    {
        setTimeout
        (
            function()
            {
                resolve();
            }, 
            miliseconds
        );
    });
}

// sleep random seconds between min and max
var asyncSleepRandom = async function(min_sec, max_sec)
{

    // generate random integer
    var random_sleep_sec = randomIntegerBetween(min_sec, max_sec);
    if(random_sleep_sec === -1)
        return;
    if(random_sleep_sec === 0)
        return;

    // sleep async
    await asyncSleep(random_sleep_sec);

}

// sleep random seconds between min and max
var asyncSleepRandomMiliseconds = async function(min_mili, max_mili)
{

    // generate random integer
    var random_sleep_mili = randomIntegerBetween(min_mili, max_mili);
    if(random_sleep_mili === -1)
        return;
    if(random_sleep_mili === 0)
        return;

    // sleep async
    await asyncSleepMiliseconds(random_sleep_mili);

}

// export utilities
exported.asyncSleep                  = asyncSleep;
exported.asyncSleepMiliseconds       = asyncSleepMiliseconds;
exported.asyncSleepRandom            = asyncSleepRandom;
exported.asyncSleepRandomMiliseconds = asyncSleepRandomMiliseconds;


// %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
// %%% Runtime Information %%%%%%%%%%%%%%%%%%%%%%%%%%%
// %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

// initialize run info
var initRunInfo = function(type)
{

    // exit if our structure is already populated
    if(isEmpty(run_info.run_id) !== true)
        return false;

    // set run information
    run_info.run_id   = guid();
    run_info.pid      = process.pid;
    run_info.type     = type;
    run_info.start_ts = Math.floor(new Date() / 1000);

    // return indicating success
    return true;

}

// export utilites
exported.run_info    = run_info;
exported.initRunInfo = initRunInfo;


// %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
// %%% Random Number Utilities %%%%%%%%%%%%%%%%%%%%%%%
// %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

// random integer between
var randomIntegerBetween = function(min, max)
{

    // run basic checks
    if(min > max)
        return -1;
    if(min === max)
        return min;

    // generate random
    return Math.floor(Math.random()*(max-min+1)+min);

}

// export the random generator
exported.randomIntegerBetween = randomIntegerBetween;


// %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
// %%% Capitalize Strings %%%%%%%%%%%%%%%%%%%%%%%%%%%%
// %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

// capitalizes words in a string
var capitalizeWords = function(str, separator = " ")
{

    if(typeof str !== "string")
        return null;
    if(str.length <= 0)
        return null;

    // try to separate string
    var separated_string = str.split(separator);
    if(isEmpty(separated_string) === true)
        return str.charAt(0).toUpperCase();
    
    // capitalize first letters
    var capitalized_string = str.split(separator).map((s) => s.charAt(0).toUpperCase() + s.substring(1)).join(separator);

    // return the capitalized string
    return capitalized_string;

}

// capitalize some words
exported.capitalizeWords = capitalizeWords;


// %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
// %%% Random String Utilities %%%%%%%%%%%%%%%%%%%%%%%
// %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

// char sets
var chars_a_to_z_small = ['a', 'b', 'c', 'd', 'e', 'f', 'g', 'h', 'i', 'j', 'k', 'l', 'm', 'n', 'o', 'p', 'q', 'r', 's', 't', 'u', 'v', 'w', 'x', 'y', 'z'];
var chars_a_to_z_caps  = ['A', 'B', 'C', 'D', 'E', 'F', 'G', 'H', 'I', 'J', 'K', 'L', 'M', 'N', 'O', 'P', 'Q', 'R', 'S', 'T', 'U', 'V', 'W', 'X', 'Y', 'Z'];
var chars_zero_to_nine = ['0', '1', '2', '3', '4', '5', '6', '7', '8', '9', '0'];
var chars_special      = 
[
    '!', '@', '#', '$', '%', '^', '&', '*',   '(',
    ')', '_', '+', '-', '=', '{', '}', '[',   ']',
    '|', '\\', ':', ';', '"', '\'', '<', '>', '?', 
    ',', '.', '/', '~', '`'
];

// all hex
var chars_all_hex_00_to_ff = 
[
    "\x00", "\x01", "\x02", "\x03", "\x04", "\x05", "\x06", "\x07", "\x08", "\x09", "\x0a", "\x0b", "\x0c", "\x0d", "\x0e", "\x0f", 
    "\x10", "\x11", "\x12", "\x13", "\x14", "\x15", "\x16", "\x17", "\x18", "\x19", "\x1a", "\x1b", "\x1c", "\x1d", "\x1e", "\x1f", 
    "\x20", "\x21", "\x22", "\x23", "\x24", "\x25", "\x26", "\x27", "\x28", "\x29", "\x2a", "\x2b", "\x2c", "\x2d", "\x2e", "\x2f", 
    "\x30", "\x31", "\x32", "\x33", "\x34", "\x35", "\x36", "\x37", "\x38", "\x39", "\x3a", "\x3b", "\x3c", "\x3d", "\x3e", "\x3f", 
    "\x40", "\x41", "\x42", "\x43", "\x44", "\x45", "\x46", "\x47", "\x48", "\x49", "\x4a", "\x4b", "\x4c", "\x4d", "\x4e", "\x4f", 
    "\x50", "\x51", "\x52", "\x53", "\x54", "\x55", "\x56", "\x57", "\x58", "\x59", "\x5a", "\x5b", "\x5c", "\x5d", "\x5e", "\x5f", 
    "\x60", "\x61", "\x62", "\x63", "\x64", "\x65", "\x66", "\x67", "\x68", "\x69", "\x6a", "\x6b", "\x6c", "\x6d", "\x6e", "\x6f", 
    "\x70", "\x71", "\x72", "\x73", "\x74", "\x75", "\x76", "\x77", "\x78", "\x79", "\x7a", "\x7b", "\x7c", "\x7d", "\x7e", "\x7f", 
    "\x80", "\x81", "\x82", "\x83", "\x84", "\x85", "\x86", "\x87", "\x88", "\x89", "\x8a", "\x8b", "\x8c", "\x8d", "\x8e", "\x8f", 
    "\x90", "\x91", "\x92", "\x93", "\x94", "\x95", "\x96", "\x97", "\x98", "\x99", "\x9a", "\x9b", "\x9c", "\x9d", "\x9e", "\x9f", 
    "\xa0", "\xa1", "\xa2", "\xa3", "\xa4", "\xa5", "\xa6", "\xa7", "\xa8", "\xa9", "\xaa", "\xab", "\xac", "\xad", "\xae", "\xaf", 
    "\xb0", "\xb1", "\xb2", "\xb3", "\xb4", "\xb5", "\xb6", "\xb7", "\xb8", "\xb9", "\xba", "\xbb", "\xbc", "\xbd", "\xbe", "\xbf", 
    "\xc0", "\xc1", "\xc2", "\xc3", "\xc4", "\xc5", "\xc6", "\xc7", "\xc8", "\xc9", "\xca", "\xcb", "\xcc", "\xcd", "\xce", "\xcf", 
    "\xd0", "\xd1", "\xd2", "\xd3", "\xd4", "\xd5", "\xd6", "\xd7", "\xd8", "\xd9", "\xda", "\xdb", "\xdc", "\xdd", "\xde", "\xdf", 
    "\xe0", "\xe1", "\xe2", "\xe3", "\xe4", "\xe5", "\xe6", "\xe7", "\xe8", "\xe9", "\xea", "\xeb", "\xec", "\xed", "\xee", "\xef", 
    "\xf0", "\xf1", "\xf2", "\xf3", "\xf4", "\xf5", "\xf6", "\xf7", "\xf8", "\xf9", "\xfa", "\xfb", "\xfc", "\xfd", "\xfe", "\xff"
];

// all hex without null byte
var chars_all_hex_01_to_ff = 
[
    "\x01", "\x02", "\x03", "\x04", "\x05", "\x06", "\x07", "\x08", "\x09", "\x0a", "\x0b", "\x0c", "\x0d", "\x0e", "\x0f", 
    "\x10", "\x11", "\x12", "\x13", "\x14", "\x15", "\x16", "\x17", "\x18", "\x19", "\x1a", "\x1b", "\x1c", "\x1d", "\x1e", "\x1f", 
    "\x20", "\x21", "\x22", "\x23", "\x24", "\x25", "\x26", "\x27", "\x28", "\x29", "\x2a", "\x2b", "\x2c", "\x2d", "\x2e", "\x2f", 
    "\x30", "\x31", "\x32", "\x33", "\x34", "\x35", "\x36", "\x37", "\x38", "\x39", "\x3a", "\x3b", "\x3c", "\x3d", "\x3e", "\x3f", 
    "\x40", "\x41", "\x42", "\x43", "\x44", "\x45", "\x46", "\x47", "\x48", "\x49", "\x4a", "\x4b", "\x4c", "\x4d", "\x4e", "\x4f", 
    "\x50", "\x51", "\x52", "\x53", "\x54", "\x55", "\x56", "\x57", "\x58", "\x59", "\x5a", "\x5b", "\x5c", "\x5d", "\x5e", "\x5f", 
    "\x60", "\x61", "\x62", "\x63", "\x64", "\x65", "\x66", "\x67", "\x68", "\x69", "\x6a", "\x6b", "\x6c", "\x6d", "\x6e", "\x6f", 
    "\x70", "\x71", "\x72", "\x73", "\x74", "\x75", "\x76", "\x77", "\x78", "\x79", "\x7a", "\x7b", "\x7c", "\x7d", "\x7e", "\x7f", 
    "\x80", "\x81", "\x82", "\x83", "\x84", "\x85", "\x86", "\x87", "\x88", "\x89", "\x8a", "\x8b", "\x8c", "\x8d", "\x8e", "\x8f", 
    "\x90", "\x91", "\x92", "\x93", "\x94", "\x95", "\x96", "\x97", "\x98", "\x99", "\x9a", "\x9b", "\x9c", "\x9d", "\x9e", "\x9f", 
    "\xa0", "\xa1", "\xa2", "\xa3", "\xa4", "\xa5", "\xa6", "\xa7", "\xa8", "\xa9", "\xaa", "\xab", "\xac", "\xad", "\xae", "\xaf", 
    "\xb0", "\xb1", "\xb2", "\xb3", "\xb4", "\xb5", "\xb6", "\xb7", "\xb8", "\xb9", "\xba", "\xbb", "\xbc", "\xbd", "\xbe", "\xbf", 
    "\xc0", "\xc1", "\xc2", "\xc3", "\xc4", "\xc5", "\xc6", "\xc7", "\xc8", "\xc9", "\xca", "\xcb", "\xcc", "\xcd", "\xce", "\xcf", 
    "\xd0", "\xd1", "\xd2", "\xd3", "\xd4", "\xd5", "\xd6", "\xd7", "\xd8", "\xd9", "\xda", "\xdb", "\xdc", "\xdd", "\xde", "\xdf", 
    "\xe0", "\xe1", "\xe2", "\xe3", "\xe4", "\xe5", "\xe6", "\xe7", "\xe8", "\xe9", "\xea", "\xeb", "\xec", "\xed", "\xee", "\xef", 
    "\xf0", "\xf1", "\xf2", "\xf3", "\xf4", "\xf5", "\xf6", "\xf7", "\xf8", "\xf9", "\xfa", "\xfb", "\xfc", "\xfd", "\xfe", "\xff"
];


// all hex from 7f to ff
var chars_all_hex_7f_to_ff = 
[
    "\x7f", 
    "\x80", "\x81", "\x82", "\x83", "\x84", "\x85", "\x86", "\x87", "\x88", "\x89", "\x8a", "\x8b", "\x8c", "\x8d", "\x8e", "\x8f", 
    "\x90", "\x91", "\x92", "\x93", "\x94", "\x95", "\x96", "\x97", "\x98", "\x99", "\x9a", "\x9b", "\x9c", "\x9d", "\x9e", "\x9f", 
    "\xa0", "\xa1", "\xa2", "\xa3", "\xa4", "\xa5", "\xa6", "\xa7", "\xa8", "\xa9", "\xaa", "\xab", "\xac", "\xad", "\xae", "\xaf", 
    "\xb0", "\xb1", "\xb2", "\xb3", "\xb4", "\xb5", "\xb6", "\xb7", "\xb8", "\xb9", "\xba", "\xbb", "\xbc", "\xbd", "\xbe", "\xbf", 
    "\xc0", "\xc1", "\xc2", "\xc3", "\xc4", "\xc5", "\xc6", "\xc7", "\xc8", "\xc9", "\xca", "\xcb", "\xcc", "\xcd", "\xce", "\xcf", 
    "\xd0", "\xd1", "\xd2", "\xd3", "\xd4", "\xd5", "\xd6", "\xd7", "\xd8", "\xd9", "\xda", "\xdb", "\xdc", "\xdd", "\xde", "\xdf", 
    "\xe0", "\xe1", "\xe2", "\xe3", "\xe4", "\xe5", "\xe6", "\xe7", "\xe8", "\xe9", "\xea", "\xeb", "\xec", "\xed", "\xee", "\xef", 
    "\xf0", "\xf1", "\xf2", "\xf3", "\xf4", "\xf5", "\xf6", "\xf7", "\xf8", "\xf9", "\xfa", "\xfb", "\xfc", "\xfd", "\xfe", "\xff"
];

// generate a random string
var randomString = function (length, chars) 
{
    var result = '';
    for (var i = length; i > 0; --i) result += chars[Math.floor(Math.random() * chars.length)];
    return result;
}


// generate a random alphabet string
var randomAlphabetString = function (length) 
{

	// set characters to use
	var chars = [].concat
	(
		chars_a_to_z_small,
		chars_a_to_z_caps
	);

	// create random string
    var result = '';
	for (var i = length; i > 0; --i) result += chars[Math.floor(Math.random() * chars.length)];
	
	// return result
    return result;
}

// generate a random string
var randomAlphanumericString = function (length) 
{

	// set characters to use
	var chars = [].concat
	(
		chars_a_to_z_small,
		chars_a_to_z_caps,
		chars_zero_to_nine
	);

	// create random string
    var result = '';
	for (var i = length; i > 0; --i) result += chars[Math.floor(Math.random() * chars.length)];
	
	// return result
    return result;
}


// export character set arrays
exported.chars_a_to_z_small     = chars_a_to_z_small;
exported.chars_a_to_z_caps      = chars_a_to_z_caps;
exported.chars_zero_to_nine     = chars_zero_to_nine;
exported.chars_special          = chars_special;
exported.chars_all_hex_00_to_ff = chars_all_hex_00_to_ff;
exported.chars_all_hex_01_to_ff = chars_all_hex_01_to_ff;
exported.chars_all_hex_7f_to_ff = chars_all_hex_7f_to_ff;

// export random string functions
exported.randomString             = randomString;
exported.randomAlphabetString     = randomAlphabetString;
exported.randomAlphanumericString = randomAlphanumericString;


// %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
// %%% ID Generation/Retrieval %%%%%%%%%%%%%%%%%%%%%%%
// %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

// create guid 
var guid = function()
{
    return uuid_random();
}

// export guid creator(s)
exported.guid = guid;


// %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
// %%% Validation Utilities %%%%%%%%%%%%%%%%%%%%%%%%%%
// %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

// checks if object, string, or array is empty.  (true/false)
//   empty types:
//           undefined
//           null
//           zero length array
//           zero length string
//           object with no keys
var isEmpty = function(test_if_this_is_empty)
{

    // undefined values are considered empty
    if(typeof test_if_this_is_empty === "undefined")
        return true;

    // null values are considered empty
    if(test_if_this_is_empty === null)
        return true;

    // null values are considered empty
    if(test_if_this_is_empty == null)
        return true;

    // zero length arrays are considered empty
    if(Array.isArray(test_if_this_is_empty) === true)
    {
        if(test_if_this_is_empty.length === 0)
            return true;
    }

    // zero length strings are considered empty
    if(typeof test_if_this_is_empty === "string")
    {
        if(test_if_this_is_empty.length === 0)
            return true;
    }

    // object with no keys is considered empty
    if(typeof test_if_this_is_empty === "object")
    {
        
        // first check if the object is a date (has the getMonth property set)
        if(typeof test_if_this_is_empty.getMonth === 'function')
        {
            return false;
        }
        else if(Object.keys(test_if_this_is_empty).length === 0)
        {
            return true;
        }
    }

    
    // the object is not empty
    return false;

}

// checks if an array of values are "empty" using the isEmpty routine.
//   empty types:
//           undefined
//           null
//           zero length array
//           zero length string
//           object with no keys
var valuesAreEmpty = function(array_of_values_to_check)
{

    // ensure it's an array
    if(Array.isArray(array_of_values_to_check) !== true)
        return true;
    
    // iterate through values
    for(var array_idx in array_of_values_to_check)
    {
        if(isEmpty(array_of_values_to_check[array_idx]) !== true)
            return false;
    }

    // return indicating success
    return true;

}

// checks that all values provided are not empty
var valuesAreNotEmpty = function(array_of_values_to_check)
{
    // ensure it's an array
    if(Array.isArray(array_of_values_to_check) === false)
        return false;

    // iterate through values
    for(var array_idx in array_of_values_to_check)
    {
        if(isEmpty(array_of_values_to_check[array_idx]) === true)
            return false;
    }

    // return indicating success
    return true;
}
// export empty check
exported.isEmpty           = isEmpty;
exported.valuesAreEmpty    = valuesAreEmpty;
exported.valuesAreNotEmpty = valuesAreNotEmpty;


// %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
// %%% Module Loader Utilities %%%%%%%
// %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

// return loaded modules
var getLoadedModules = async function()
{
    return app_config.dependencies;
}

// Load a module (ptusa specific replacement for require() which allows us
// to load modules using the app config passed into this module itself).  Modules
// are loaded and returned, but they are also stored within the app config dependencies
// object to prevent duplicate initializations.
var loadModule = async function(module_path, export_name = null, dependency_set = "misc", attempt_to_instantiate = false)
{

    // ensure app config is loaded
    if(isEmpty(app_config) === true)
    {
        console.log("Err: Cannot loadModule() without a valid app_config.");
        process.exit(0);
        return;
    }

    // create dependencies if necessary
    if(isEmpty(app_config.dependencies) === true)
        app_config.dependencies = {};
    if(isEmpty(app_config.dependencies[dependency_set]) === true)
        app_config.dependencies[dependency_set] = {};

    // skip loading the module if it's already loaded
    if(isEmpty(app_config.dependencies[dependency_set][export_name]) === false)
    {
        // console.log({module_already_loaded: module_path})
        return app_config.dependencies[dependency_set][export_name];
    }

    // require the module
    var mod = null;
    try
    {
        mod = require(module_path);
    }
    catch(err)
    {
        console.log
        ({
            could_not_load_module: module_path,
            message: err.message
		});
		console.log(err.stack);
        process.exit(0);
    }
    
    // ensure we have an init function for the module
    if(typeof mod.init != "function")
    {
        console.log("Error: Loaded module does not have an init() method.");
        process.exit(0);
    }

    // initialize the module
    var loaded_mod = null;
    if(typeof mod.init === "function")
    {
        try
        {
            loaded_mod = await mod.init(app_config);
        }
        catch(error)
        {
            console.log
            ({
                failed_to_load_module: "Module could not be loaded due to error.",
                err_message: error.message,
                module_path: module_path
            });
            console.log(error.stack);
            process.exit(0);
        }
    }

    // if we have an export name, provide export
    if(export_name !== null)
    {

        if(isEmpty(loaded_mod[export_name]) === true)
        {
            notice
            ({
                loadModule: "Failed to load module.",
                export_name: export_name
            });
        }

        // add export to the dependency set
        app_config.dependencies[dependency_set][export_name] = loaded_mod[export_name];

        // attempt to create instance of the class, then delete it (this creates class documentation)
        if(attempt_to_instantiate === true && export_name !== "DocumentationGenerator")
        {
            
            try
            {
                var test = new app_config.dependencies[dependency_set][export_name]({});
            } catch(err){}

            try
            {
                var test = new app_config.dependencies[dependency_set][export_name]();
            } catch(err){}
        }

        // console.log({loaded_mod: loaded_mod});

        // return the loaded mod
        return loaded_mod[export_name];

    }
    
    // return the loaded module
    return loaded_mod;

}

// load a module from audit cluster
var loadAuditClusterModule = async function(audit_cluster_relative_module_path, export_name = null)
{
        
    // attempt to load the module
    var loaded_module = await loadModule
    (
        app_config.shared_library_path + "ptusa_audit_cluster_shared_libs/custom/" + audit_cluster_relative_module_path,
        export_name,
        "audit_cluster"
    );   

    // return the loaded module
    return loaded_module;

}

// load an audit cluster job module
var loadAuditClusterJobModule = async function(audit_cluster_relative_module_path, export_name = null)
{
        
    // attempt to load the module
    var loaded_module = await loadModule
    (
        app_config.shared_library_path + "ptusa_audit_cluster_shared_libs/custom/" + audit_cluster_relative_module_path,
        export_name,
        "audit_cluster_job"
    );   

    // return the loaded module
    return loaded_module;

}

// load a module from the pentest tool
var loadPentestToolModule = async function(pentest_tool_relative_module_path, export_name = null)
{

    // attempt to load the module
    var loaded_module = await loadModule
    (
        app_config.shared_library_path + "ptusa_pentest_tool_shared_libs/custom/" + pentest_tool_relative_module_path, 
        export_name,
        "pentest_tool"
    ); 

    // return the loaded module
    return loaded_module;

}

// load a custom pentest report generator module
var loadPentestToolCustomReportGeneratorModule = async function(pentest_tool_full_module_path, export_name = null)
{

    var loaded_module = null;
    
    // attempt to load the module
    var loaded_module = await loadModule
    (
        pentest_tool_full_module_path, 
        export_name,
        "pentest_tool"
    ); 
    

    // return the loaded module
    return loaded_module;

}

// load a module from the web analyzer tool
var loadWebAnalyzerToolModule = async function(web_analyzer_tool_relative_module_path, export_name = null)
{

    // attempt to load the module
    var loaded_module = await loadModule
    (
        app_config.shared_library_path + "ptusa_web_analysis_tool_shared_libs/custom/" + web_analyzer_tool_relative_module_path, 
        export_name,
        "web_analyzer_tool"
    ); 

    // return the loaded module
    return loaded_module;

}

// export audit cluster module loader utilities
exported.loadModule                = loadModule;
exported.loadAuditClusterModule    = loadAuditClusterModule;
exported.loadPentestToolModule     = loadPentestToolModule;
exported.loadPentestToolCustomReportGeneratorModule = loadPentestToolCustomReportGeneratorModule;
exported.getLoadedModules          = getLoadedModules;
exported.loadWebAnalyzerToolModule = loadWebAnalyzerToolModule;


// %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
// %%% File Related Validators %%%%%%%%%%%%%%%%%%%%%%%
// %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

// Basic check to see if a filename is sensible or not.
var filenameIsSensible = function(filename, results_in_out = null, hidden_files_ok = false)
{

    // ensure it's a string and has a length
    if(typeof filename !== "string")
    {
        if(results_in_out !== null)
            results_in_out.is_not_string = true;
        return false;
    }
    if(filename.length === 0)
    {
        if(results_in_out !== null)
            results_in_out.is_zero_length_string = true;
        return false;
    }

    // check for hidden prefix
    if(hidden_files_ok === false)
    if(filename[0] === '.')
    {
        if(results_in_out !== null)
            results_in_out.contains_hidden_file_dot_prefix = true;
        return false;
    }

    // test that the file doesn't contain double dot sequences
    if(/\.\./.test(filename) === true)
    {
        if(results_in_out !== null)
            results_in_out.contains_dot_dot_sequence = true;
        return false;
    }

    // regexp for allowable characters
    var allowable_chars_regex = /^[0-9a-zA-Z\._-]*$/;

    // test that the filename only has allowable characters
    if(allowable_chars_regex.test(filename) !== true)
    {
        if(results_in_out !== null)
            results_in_out.contains_invalid_characters = true;
        return false;
    }

    // return indicating success
    return true;

}

// check if a directory exists
var directoryExists = function(directory)
{
    return fs.existsSync(directory);
}

// check if a file exists
var fileExists = function(full_file_path)
{
    return fs.existsSync(full_file_path);
}

// export file/directory utilities
exported.filenameIsSensible = filenameIsSensible;
exported.directoryExists    = directoryExists;
exported.fileExists         = fileExists;


// %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
// %%% Object Search/Set Utiliies %%%%%%%%%%%%%%%%%%%%
// %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

// Searches an object using a path array and returns an object property.  Returns null
// if a property does not exist.
var getObjectElementByPath = function(object_to_search, path_array)
{

    // ensure we have search items
    if(isEmpty(object_to_search) === true)
        return null;
    if(Array.isArray(path_array) === false)
        return null;

    // return nested keys
    return nested_keys.get(object_to_search, path_array);

}

// Sets an object using a path array and returns an object property.  Returns false
// if a property does not exist and cannot be assigned
var setObjectElementByPath = function(object_to_search, path_array, value_to_assign)
{

    // ensure we have search items
    if(isEmpty(object_to_search) === true)
        return false;
    if(Array.isArray(path_array) === false)
        return false;

    // return nested keys
    return nested_keys.set(object_to_search, path_array, value_to_assign);

}

// export utilities
exported.getObjectElementByPath = getObjectElementByPath;
exported.setObjectElementByPath = setObjectElementByPath;


// %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
// %%% Load File As Base64 %%%%%%%%%%%%%%%%%%%%%%%%%%%
// %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

// attempt to load file as base64
var loadFileAsBase64 = function(full_file_path)
{

    if(isEmpty(full_file_path) === true)
        return null;
    if(fs.existsSync(full_file_path) !== true)
        return null;

    // attempt to read file data
    var file_data = fs.readFileSync(full_file_path);

    // ensure we have file data
    if(isEmpty(file_data) === true)
        return null;

    // set ret data
    var ret_data = file_data.toString('base64');

    // return the ret data
    return ret_data;

}

// load file as png
exported.loadFileAsBase64 = loadFileAsBase64;


// %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
// %%% Load File As Array %%%%%%%%%%%%%%%%%%%%%%%%%%%%
// %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

// read file lines into an array (skips empty lines)
var readFileLinesIntoArray = function(filename)
{

    // gather data
    var fs = require('fs');
    var ret_array = null;
    try
    {
        notice(filename);
        ret_array = fs.readFileSync(filename).toString().split("\n");
    } catch(err){}

    if(isEmpty(ret_array) === true)
    {
        try
        {
            notice(filename);
            ret_array = fs.readFileSync(filename).toString().split("\r");
        } catch(err){}
    }

    // final array
    var final_array = [];

    // iterate through array
    for(i in ret_array) 
    {

        // set element from array
        var element = ret_array[i];

        // remove any newlines or returns that may have somehow got in
        element = element.replace(/\n$/, '');
        element = element.replace(/\r$/, '');
        
        // ensure the element isn't empty
        if(isEmpty(element) === true)
            continue;
        
        // push final array
        final_array.push(element);

    }

    // return the file array
    return final_array;

}

// attempt to read an array containing file names into a single buffer with no duplicates
var readFileArrayLinesIntoArray = function(file_array)
{

    // set final array
    var final_array = [];
    if(isEmpty(file_array) === true)
        return false;
    if(Array.isArray(file_array) === false)
        return false;

    // iterate through file array
    for(var idx = 0; idx < file_array.length; idx++)
    {

        // read lines
        var line_array = readFileLinesIntoArray(file_array[idx]);
        if(isEmpty(line_array) === true)
            continue;

        // iterate through loaded line array
        for(var line_idx in line_array)
        {

            // ensure the element doesn't already exist
            if(final_array.indexOf(line_array[line_idx]) !== -1)
                continue;

            // add line if it doesn't exist in array
            final_array.push(line_array[line_idx]);

        }

    }

    // return the final array
    return final_array;

}

// set exports
exported.readFileLinesIntoArray      = readFileLinesIntoArray;
exported.readFileArrayLinesIntoArray = readFileArrayLinesIntoArray;


// %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
// %%% Image Utilities %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
// %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

// load png file
var loadPNGFile = async function(full_file_path)
{

    // run basic checks
    if(isEmpty(full_file_path) === true)
        return null;
    if(fs.existsSync(full_file_path) !== true)
        return null;

    // attempt to read file data
    var file_data = fs.readFileSync(full_file_path);

    // ensure we have file data
    if(isEmpty(file_data) === true)
        return null;

    // set ret data
    var ret_data = await parsePng(file_data);
    if(isEmpty(ret_data) === true)
        return null;

    // set base64 data
    ret_data.base64_data = loadFileAsBase64(full_file_path);

    // set parser to null (deref)
    ret_data._parser      = null;
    ret_data._packer      = null;
    ret_data._events      = null;
    ret_data._eventsCount = null;

    // remove internal references
    delete ret_data.data;
    delete ret_data._parser;
    delete ret_data._packer;
    delete ret_data._events;
    delete ret_data._eventsCount;

    // set marker as indicator
    ret_data.is_png = true;

    // return the ret data
    return ret_data;

}

// load png file
exported.loadPNGFile = loadPNGFile;


// %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
// %%% Raw JWT Decoder %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
// %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

// decode jwt
var decodeJWTFromBase64 = function(base64_jwt)
{


    if(typeof base64_jwt !== "string")
        return null;

    // run arbitrary checks
    if(base64_jwt.length > 1024 || base64_jwt.length < 20)
        return null;

    // return null if its
    if(isEmpty(base64_jwt) === true)
        return null;

    // return null if we cant decode
    var buff = Buffer.from(base64_jwt, 'base64');
    if(isEmpty(buff) === true)
        return null;

    // convert to string
    var jwt_str = buff.toString();
    
    // ensure that parts are filled out
    var parts = jwt_str.split("}");
    if(parts.length < 3)
        return null;


    // set header part
    parts[0] = parts[0]+"}";
    if(parts[0].length > 100)
        return null;

    // set data part
    parts[1] = parts[1]+"}";
    if(parts[1].length > 500)
        return null;


    var decoded = 
    {
        header: JSON.parse(parts[0]),
        body:   JSON.parse(parts[1]),
        signature: buff.slice(parts[0].length + parts[1].length)
    };

    // return decoded values
    return decoded;

}

// export jwt
exported.decodeJWTFromBase64 = decodeJWTFromBase64;


// %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
// %%% Stacktrace %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
// %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

// create stacktrace array (simplified)
var stacktrace = function()
{
    
    // create stacktrace (required for)
    var internal_stacktrace = StackTrace.getSync();

    // stack trace
    var ret_stacktrace = [];

    // create simplified stack trace
    for(var i = 0; i < internal_stacktrace.length; i++)
    {
        var stack_line = internal_stacktrace[i].fileName + ":" + internal_stacktrace[i].lineNumber + ":" + internal_stacktrace[i].functionName;
        ret_stacktrace.push(stack_line);
    }

    // return the simplified stack trace
    return ret_stacktrace;

}

// export stacktrace utility
exported.stacktrace = stacktrace;


// %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
// %%% Logging/Warn/ Error Utilities %%%%%%%%%%%%%%%%%
// %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

// marker for whether or not db has been initialized
var log_db_has_been_initialized = false;

// add log message to db
var initLogDB = async function()
{

	// THIS FUNCTION HAS BEEN DEPRECATED DUE TO SQLITE BEING AN
	// INADEQUATE SOLUTION.

    // skip if we've already initialized the db
    if(log_db_has_been_initialized === true)
        return;

    var log_db_file = app_config.logging.log_dir + 
                        "/" +
                        app_config.logging.log_file;

    // open logger if it's not open
    if(logger_db === null)
    {
		/*
        logger_db = await sqlite.open
        (
            app_config.logging.log_dir + 
            "/" +
            app_config.logging.log_file
        )
        .catch((err) =>
        {
            console.log("Failed to open database: "+log_db_file);
            console.log(err);
            process.exit(1);
		});
		*/
    }

	/*
    // ensure the db is open
    if(logger_db.driver.open !== true)
    {

        console.log("CRITICAL ERROR: cannot init log database.".red);
        process.exit(1);
        return false;
    }

    // run create table if not exists query
    await logger_db.all
    (
        `CREATE TABLE IF NOT EXISTS log 
        (
            id               integer PRIMARY KEY AUTOINCREMENT,
            process_type     text,
            process_start_ts integer,
            run_guid         text, 
            process_id       integer, 
            log_type         text, 
            message          text, 
            stacktrace       text, 
            create_ts        integer
        )` 
    );
	*/

    // log init 
    console.log("[+] Init logDB OK!: ".green+ "( pid: "+process.pid+", type: "+run_info.type+" ) '"+log_db_file+"'");

    // init marker
    log_db_has_been_initialized = true;

    // return indicating success
    return true;

}


// add log message to db
var addMessageToLogDB = async function(log_type, log_level, log_message, stacktrace)
{

	// THIS FUNCTION HAS BEEN DEPRECATED DUE TO SQLITE BEING AN
	// INADEQUATE SOLUTION.

	/*
    // init the db
    await initLogDB();

    try{

    // check if values are empty
    if(valuesAreNotEmpty([log_type, log_level, stacktrace]) === false)
    {
        console.log("[!!] addMessageToLogDB: ".red + "one or more parameters were empty.");
        console.log
        ({
            log_type:   log_type, 
            log_level:  log_level, 
            stacktrace: stacktrace
        });
        process.exit(1);
        return false;
    }

	
    // open logger if it's not open
    if(logger_db === null)
    {

        // db open success marker
        var db_opened_ok = false;

        var logger_db_file =  app_config.logging.log_dir + "/" + app_config.logging.log_file;

        // loop to ensure we opened the db ok
        while(db_opened_ok === false)
        {
            try
            {
                console.log("Attempting to open logger db".red + ": "+logger_db_file);
                logger_db = await sqlite.open(logger_db_file);
            }
            catch(err)
            {
                console.log
                ({
                    init_log_db_failed: "Opening file failed.",
                    message: err.message,
                    app_config: app_config
                });
            }

            // check that db opened ok
            if(typeof logger_db.driver === "object")
            {
                if(logger_db.driver.open === true)
                    db_opened_ok = true;
            }

            // sleep for a second if the db wasn't opened ok and try again
            if(db_opened_ok === false)
                await asyncSleep(1);

        }

    }

    // ensure the db is open
    if(logger_db.driver.open !== true)
    {
        console.log("[!!] CRITICAL ERROR: ".red + "addMessageToLogDB failed because log db could not be opened.");
        console.log(logger_db);
        process.exit(1);
        return false;
    }


    var insert_query_executed_ok = false;
    while(insert_query_executed_ok === false)
    {
        try
        {
            // run insert query
            await logger_db.all
            (
                `INSERT INTO log 
                (
                    process_type,
                    process_start_ts,
                    run_guid, 
                    process_id, 
                    log_type, 
                    message, 
                    stacktrace, 
                    create_ts
                ) 
                VALUES 
                (
                    ?,
                    ?,
                    ?,
                    ?, 
                    ?, 
                    ?, 
                    ?, 
                    ?
                )`, 
                [
                    run_info.type,
                    run_info.start_ts,
                    run_info.run_id,
                    process.pid,
                    log_type,
                    log_message,
                    JSON.stringify(stacktrace),
                    Math.floor(new Date() / 1000)
                ]
            );

            // mark query as executing ok
            insert_query_executed_ok = true;

        } 
        catch(err)
        {

            // sleep for one second if db is busy
            await asyncSleep(1);

        }
    }
	*/
    // return indicating success
    return true;

}


// notice
var notice = async function(message, trace_depth_start = 1, log_level = 1)
{
    
    // create stacktrace (required for)
    var stacktrace = StackTrace.getSync();

    // type data
    var type_data = typeof message;

    // convert message if necessary
    if(typeof message === "object" || Array.isArray(message) === true)
    {
        message = JSON.stringify(message, decycle(), 2);
    }


    if(typeof stacktrace === "object" && stacktrace !== null)
    {
        var err_trace = JSON.stringify(stacktrace, null, 2);


        // set function name if necessary
        if(typeof stacktrace[trace_depth_start] === "undefined")
            trace_depth_start = 0;
        else if(typeof stacktrace[trace_depth_start] === "undefined")
        {
            console.log("CRITICAL ERROR:".red + " Stacktrace was empty.");
            return false;
        }
        else if(typeof stacktrace[trace_depth_start].functionName === "undefined")
            stacktrace[trace_depth_start].functionName = "undefined/anonymous function";
        

        // add file/class/line number
        var log_info = "\n[Notice]-> ".green + stacktrace[trace_depth_start].fileName + ":" + stacktrace[trace_depth_start].lineNumber + "\n" +
        stacktrace[trace_depth_start].functionName.underline  + ": "+ type_data + "\n" +
        message + "\n" +  "[Notice Above]-> ".green + stacktrace[trace_depth_start].fileName + ":" + stacktrace[trace_depth_start].lineNumber + "\n";

    }
    else
    {
        log_info = "\n[Notice]-> ".green + ": " + message + "\n" +  "[Notice Above]-> ".green + stacktrace[trace_depth_start].fileName + ":" + stacktrace[trace_depth_start].lineNumber + "\n";
    }

    // display the message
    console.log(log_info);
    
}

// log an error
var err = async function(message, exit_process_on_error = false, trace_depth_start = 1, log_level = 1)
{

    exit_process_on_error = false;

    // create stacktrace (required for)
    var stacktrace = StackTrace.getSync();

    // type data
    var type_data = typeof message;
    
    // convert message if necessary
    if(typeof message === "object" || Array.isArray(message) === true)
    {
        message = JSON.stringify(message, decycle(), 2);
    }

    // this holds log information    
    var log_info = null;

    // process stack trace
    if(typeof stacktrace === "object" && stacktrace !== null)
    {

        // set function name if necessary
        if(typeof stacktrace[trace_depth_start] === "undefined")
            trace_depth_start = 0;
        else if(typeof stacktrace[trace_depth_start] === "undefined")
        {
            console.log("CRITICAL ERROR:".red + " Stacktrace was empty.");
            return false;
        }
        else if(typeof stacktrace[trace_depth_start].functionName === "undefined")
            stacktrace[trace_depth_start].functionName = "undefined/anonymous function";

    
        // add file/class/line number
        log_info = "\n[Err]-> ".red + stacktrace[trace_depth_start].fileName + ":" + stacktrace[trace_depth_start].lineNumber + "\n" +
        stacktrace[trace_depth_start].functionName.underline  + ": "+ type_data + "\n" +
        message + "\n\n"+
        "  Stack Trace:".red;

        // add full trace
        for(var trace_idx in stacktrace)
        {
            var trace_iter = stacktrace[trace_idx];
            log_info += "\n    "+trace_iter.functionName + ": "+trace_iter.fileName+ ":" + trace_iter.lineNumber;
        }

        // set log info
        log_info += "\n";

    }
    else
    {
        log_info = "\n[Err]-> ".red + ": " + message + "\n";
    }

    // add process exit on failure
    if(exit_process_on_error === true)
    {
        log_info += "\n[!!] Exiting process due to error. (exit_process_on_error === true)".red;
    }

    // display the message
    console.log(log_info);
    
    // add message to database
    await addMessageToLogDB("ERROR", log_level, message, stacktrace);

    // exit process if configured to do so
    if(exit_process_on_error === true)
        process.exit(1);

}

// log a warning
var warn = async function(message, trace_depth_start = 1, log_level = 1)
{

    // create stacktrace (required for)
    var stacktrace = StackTrace.getSync();

    // type data
    var type_data = typeof message;
    
    // convert message if necessary
    if(typeof message === "object" || Array.isArray(message) === true)
    {
        message = JSON.stringify(message, decycle(), 2);
    }

    if(typeof stacktrace === "object" && stacktrace !== null)
    {
        
        // set function name if necessary
        if(typeof stacktrace[trace_depth_start] === "undefined")
            trace_depth_start = 0;
        else if(typeof stacktrace[trace_depth_start] === "undefined")
        {
            console.log("CRITICAL ERROR:".red + " Stacktrace was empty.");
            return false;
        }
        else if(typeof stacktrace[trace_depth_start].functionName === "undefined")
            stacktrace[trace_depth_start].functionName = "undefined/anonymous function";

        
        // add file/class/line number
        var log_info = "\n[Warn]-> ".yellow + stacktrace[trace_depth_start].fileName + 
                        ":" + 
                        stacktrace[trace_depth_start].lineNumber + 
                        "\n" +
                        stacktrace[trace_depth_start].functionName.underline  + 
                        ": "+ 
                        type_data + 
                        "\n" +
                        message + "\n";

    }
    else
    {
        log_info = "\n[Warn]-> ".yellow + ": " + message + "\n";
    }
    
    // add message to database
    await addMessageToLogDB("WARNING", log_level, message, stacktrace);

    // display the message
    console.log(log_info);

}

// log a notice
var log = async function(message, trace_depth_start = 1, log_level = 1)
{

    // create stacktrace (required for)
    var stacktrace = StackTrace.getSync();

    // type data
    var type_data = typeof message;

    // convert message if necessary
    if(typeof message === "object" || Array.isArray(message) === true)
    {
        message = JSON.stringify(message, decycle(), 2);
    }

	// display the trace message to console
    if(typeof stacktrace === "object" && stacktrace !== null)
    {
        
        // set function name if necessary
        if(typeof stacktrace[trace_depth_start] === "undefined")
            trace_depth_start = 0;
        else if(typeof stacktrace[trace_depth_start] === "undefined")
        {
            console.log("CRITICAL ERROR:".red + " Stacktrace was empty.");
            return false;
        }
        else if(typeof stacktrace[trace_depth_start].functionName === "undefined")
            stacktrace[trace_depth_start].functionName = "undefined/anonymous function";
        

        // add file/class/line number
        var log_info = "\n[Log]-> ".green + stacktrace[trace_depth_start].fileName + ":" + stacktrace[trace_depth_start].lineNumber + "\n" +
        stacktrace[trace_depth_start].functionName.underline  + ": "+ type_data + "\n" +
        message + "\n";

    }
    else
    {
        log_info = "\n[Log]-> ".green + ": " + message + "\n";
    }

    // add message to database
    await addMessageToLogDB("LOG", log_level, message, stacktrace);

    // display the message
    console.log(log_info);
    

}

// error warning and log utilities
exported.initLogDB = initLogDB;
exported.err       = err;
exported.warn      = warn;
exported.log       = log;
exported.notice    = notice;


// %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
// %%% Redactions %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
// %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

// redact object by keys
var redactObjectsByKeys = function(object_to_redact, keys_to_redact)
{

    for(var key_idx in object_to_redact)
    {

        // recurse if this is an object
        if(typeof object_to_redact[key_idx] === "object" || Array.isArray(object_to_redact[key_idx]))
        {

            // see if the current value is in the list of redadction keys
            var redact_idx = keys_to_redact.indexOf(key_idx);
            if(redact_idx === -1)
            {
                redactObjectsByKeys(object_to_redact[key_idx], keys_to_redact);
            }
            else
            {
                // overwrite the object_to_redact if necessary
                object_to_redact[key_idx] = "This value has been redacted by developer.";
            }

            
        }
        else if(typeof object_to_redact[key_idx] === "string")
        {

            // see if the current value is in the list of redadction keys
            var redact_idx = keys_to_redact.indexOf(key_idx);
            if(redact_idx === -1)
                continue;
            
            // overwrite the object_to_redact if necessary
            object_to_redact[key_idx] = "This value has been redacted by developer.";

        }

    }

}


// redact object by object_to_redact
var redactObjectsByValues = function(object_to_redact, values_to_redact)
{

    for(var value_idx in object_to_redact)
    {

        // recurse if this is an object
        if(typeof object_to_redact[value_idx] === "object" || Array.isArray(object_to_redact[value_idx]))
        {
            redactObjectsByValues(object_to_redact[value_idx], values_to_redact);
        }
        else if(typeof object_to_redact[value_idx] === "string")
        {

            // see if the current value is in the list of redadction keys
            var redact_idx = values_to_redact.indexOf(object_to_redact[value_idx]);
            if(redact_idx === -1)
                continue;
            
            // overwrite the object_to_redact if necessary
            object_to_redact[value_idx] = "This value has been redacted by developer.";

        }

    }

}

// substitute object values by regex
var substituteValueInObjectByRegex = function(object_to_redact, regex_to_substitute, value_to_replace_with)
{

    // iterate object
    for(var value_idx in object_to_redact)
    {

        // recurse if this is an object
        if(typeof object_to_redact[value_idx] === "object" || Array.isArray(object_to_redact[value_idx]))
        {
            substituteValueInObjectByRegex(object_to_redact[value_idx], regex_to_substitute, value_to_replace_with);
        }
        else if(typeof object_to_redact[value_idx] === "string")
        {

            // see if the current value is in the list of redadction keys
            var matched_regex = regex_to_substitute.test(object_to_redact[value_idx]);
            if(matched_regex === false)
                continue;

            // perform replacement
            object_to_redact[value_idx] = object_to_redact[value_idx].replace(regex_to_substitute, value_to_replace_with);

        }

    }

}

// redact objects
exported.redactObjectsByValues = redactObjectsByValues;
exported.redactObjectsByKeys   = redactObjectsByKeys;

// replace in objects by regex
exported.substituteValueInObjectByRegex = substituteValueInObjectByRegex;




// %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
// %%% Check Port Lists %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
// %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

// check port number or range
var checkNmapPortNumberOrRange = function(port_number_or_range)
{

    // ensure port list isn't empty
    if(isEmpty(port_number_or_range) === true)
        return false;
    if(typeof port_number_or_range !== "string")
        return false;

    // run initial regex match
    if(/[\-0-9]+$/.test(port_number_or_range) === false)
    {
        return false;
    }

    // check if we have a range
    if(port_number_or_range.indexOf('-') === -1)
    {

        // parse the port number
        var port_number = parseInt(port_number_or_range);
        if(Number.isInteger(port_number) === false)
            return false;

        // ensure port is within reasonable port range
        if(port_number > 65534)
            return false;
        if(port_number < 0)
            return false;

        // checks passed ok
        return true;

    }
    else
    {

        // gather the port range
        var port_range = port_number_or_range.split('-');
        if(port_range.length !== 2)
            return false;

        // gather bottom and top
        var port_bottom = port_range[0];
        var port_top    = port_range[1];

        // parse values
        port_bottom = parseInt(port_bottom);
        port_top    = parseInt(port_top);

        // check values
        if(port_bottom === port_top)
            return false;
        if(port_bottom > port_top)
            return false;

        // ensure port_top is within reasonable port range
        if(port_bottom > 65534)
            return false;
        if(port_bottom < 0)
            return false;

        // ensure port_top is within reasonable port range
        if(port_top > 65534)
            return false;
        if(port_top < 0)
            return false;

        // checks passed ok
        return true;

    }
}


// Ensure nmap port list only contains reasonable entries.  This is used by
// jobs in order to check port lists don't contain weird formats.
var checkNmapPortListStringIsValid = function(port_list)
{

    // ensure port list isn't empty
    if(isEmpty(port_list) === true)
        return false;
    if(typeof port_list !== "string")
        return false;

    // run initial regex match
    if(/[,\-0-9]+$/.test(port_list) === false)
    {
        return false;
    }

    // check if we have a comma or not
    if(port_list.indexOf(',') === -1)
    {

        // check number or port range
        return checkNmapPortNumberOrRange(port_list);

    }

    // split the list
    var split_list = port_list.split(',');

    // iterate through split entries
    for(var idx = 0; idx < split_list.length; idx++)
    {

        // port set
        var port_set = split_list[idx];

        // check number or port range
        var passed_check = checkNmapPortNumberOrRange(port_set);
        if(passed_check === false)
            return false;

    }

    // return indicating that port list is valid
    return true;

}

// export the routine
exported.checkNmapPortListStringIsValid = checkNmapPortListStringIsValid;


// %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
// %%% Split String Based On Criteria %%%%%%%%%%%%%%%%%%
// %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

// split a string by length
var splitStringByLength = function(string, length, replacement = "\n")
{

    // ensure we have a string
    if(isEmpty(string) === true)
        return "";
    if(isEmpty(length) === true)
        return "";
    if(length <= 0)
        return "";

    var new_string = "";

    var length_counter = 0;
    for(var idx = 0; idx < string.length; idx++)
    {

        // append character to string
        new_string += string[idx] + "";

        // break string where necessary
        if(length_counter === length)
        {
            new_string += replacement;
            length_counter = 0;
        }
        else
        {
            length_counter++;
        }
    }

    return new_string;
}

// split the string by length
exported.splitStringByLength = splitStringByLength;


// %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
// %%% CIDR Utilities %%%%%%%%%%%%%%%%%%%%%%%%%%
// %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

// returns a list of cidr entries from a given range (be careful with large ranges)
var getIPArrayFromCIDR = function(cidr_range)
{

    // run empty check
    if(isEmpty(cidr_range) === true)
        return null;

    var cidr = new ip_cidr(cidr_range);
    if(cidr.isValid() !== true)
        return null;

    // gather the address array
    var address_array = cidr.toArray();

    // return the array
    return address_array;

}

// export the cidr utility
exported.getIPArrayFromCIDR = getIPArrayFromCIDR;


// %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
// %%% Port Utilities %%%%%%%%%%%%%%%%%%%%%%%%%%
// %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

// gathers netstat information
var getNetstat = async function(params)
{

    // set filter if one is provided
    var filter = null;
    try
    {
        filter = params.filter;
    } catch(err){}

    var port_states = [];

    var local_ports = 
    {
        tcp:  {},
        udp:  {},
        tcp6: {},
        udp6: {}
    };

    var remote_ports = 
    {
        tcp: {},
        udp: {},
        tcp6: {},
        udp6: {}
    };

    var listening_ports = 
    {
        tcp: {},
        udp: {},
        tcp6: {},
        udp6: {}
    };

    /*
    // ensure that protocol is within 
    switch(params.protocol)
    {
        case "tcp":
        case "tcp6":
        case "udp":
            break;
        default:
            return null;
    };
    */

    await new Promise(function(resolve, reject)
    {
        netstat
        ({
            filter: filter,
            done: function(error)
            {
                console.log("Netstat finished.");
                resolve();
            }
        }, function(data)
        {

            // add remote port entry
            if(data.remote.port !== null && isNaN(data.remote.port) === false)
            {
                
                // ensure we have an object entry
                if(isEmpty(remote_ports[data.protocol]) === true)
                    remote_ports[data.protocol] = {};

                // add listener entry if this is a listening port
                if(data.state === "LISTEN")
                {

                    if(isEmpty(listening_ports[data.protocol]) === true)
                        listening_ports[data.protocol] = {};

                    // add port data
                    listening_ports[data.protocol][data.remote.port] = JSON.parse(JSON.stringify(data));

                }

                // add port data
                remote_ports[data.protocol][data.remote.port] = JSON.parse(JSON.stringify(data));

            }
            
            // add local port entry
            if(data.local.port !== null && isNaN(data.local.port) === false)
            {

                // ensure we have an object entry
                if(isEmpty(remote_ports[data.protocol]) === true)
                    local_ports[data.protocol] = {};

                if(data.state === "LISTEN")
                {
                    if(isEmpty(listening_ports[data.protocol]) === true)
                        listening_ports[data.protocol] = {};

                    // set port entry
                    listening_ports[data.protocol][data.local.port] = JSON.parse(JSON.stringify(data));

                }

                // add port data
                local_ports[data.protocol][data.local.port] = JSON.parse(JSON.stringify(data));

            }
            
            
        });
    })

    // set port states
    var port_states = 
    {
        remote_ports: remote_ports,
        local_ports:  local_ports,
        listening_ports: listening_ports
    };

    // return the lookup data
    return port_states;

}

// check if port is listening
var networkPortIsListening = async function(params)
{

    if(isEmpty(params.protocol) === true)
        return null;
    if(isEmpty(params.port) === true)
        return null;
    
    // gather netstat output
    var netstat_output = await getNetstat
    ({
        filter:
        {
            protocol: params.protocol,
            local:
        {
            port: params.port
        },
            state: "LISTEN"
        }
    });

    // check output to see if port is listening
    var port_is_listening = false;
    try
    {

        if(netstat_output.listening_ports[params.protocol][params.port].state === "LISTEN")
            port_is_listening = true;

    } catch(err){}

    // return the netstat output
    return port_is_listening;

}

// This retrieves an unused local bind port which can be used
// to bind a service.
var getUnusedLocalTCPListenerPort = async function()
{

    // this holds the unused port
    var unused_port = null;

    // iterate till we find a port we can bind on
    var port_is_listening = true;
    while(port_is_listening === true)
    {

        // generate a random ephemerial port
        unused_port = randomIntegerBetween(1025, 65533);

        // check if the port is listening
        var port_is_listening = await networkPortIsListening
        ({
            protocol: "tcp",
            port:     unused_port
        });

        
    }
    
    // return the unused port
    return unused_port;

}

// export utilities
exported.getNetstat                    = getNetstat;
exported.networkPortIsListening        = networkPortIsListening;
exported.getUnusedLocalTCPListenerPort = getUnusedLocalTCPListenerPort;


// %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
// %%% Process Exit Utilities %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
// %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

// Exit the process, with stack tracing.
var exit = async function(message)
{
        
    var trace_depth_start = 1;
    var log_level = 1;

	// this determines if we're in process, or in a thread
	var process_or_thread = "Process";

	// sleep a sec if exiting from a worker thread
	if(!isMainThread)
	{
		process_or_thread = "WorkerThread";
	}

    // create stacktrace (required for)
    var stacktrace = StackTrace.getSync();

    // type data
    var type_data = typeof message;

    // convert message if necessary
    if(typeof message === "object" || Array.isArray(message) === true)
    {
        message = JSON.stringify(message, decycle(), 2);
    }

    // log information
    var log_info = "";

	// if we have a stacktrace, use it
    if(typeof stacktrace === "object" && stacktrace !== null)
    {

		// gather stack trace
        var err_trace = JSON.stringify(stacktrace, 1, 2);

        // set function name if necessary
        if(typeof stacktrace[trace_depth_start] === "undefined")
            trace_depth_start = 0;
        else if(typeof stacktrace[trace_depth_start] === "undefined")
        {
            console.log("CRITICAL ERROR:".red + " Stacktrace was empty.");
            return false;
        }
        else if(typeof stacktrace[trace_depth_start].functionName === "undefined")
            stacktrace[trace_depth_start].functionName = "undefined/anonymous function";
		
		// set top exit string
		var exiting_proess_or_thread_string = "\n[Exiting "+process_or_thread+"]-> ";
		exiting_proess_or_thread_string = exiting_proess_or_thread_string.red;

		// set bottom exit string
		var exiting_process_or_thread_string_below = "\n[Exiting "+process_or_thread+" Above]-> ";
		exiting_process_or_thread_string_below = exiting_process_or_thread_string_below.red;

        // add file/class/line number
        log_info = exiting_proess_or_thread_string.red + stacktrace[trace_depth_start].fileName + ":" + stacktrace[trace_depth_start].lineNumber + "\n" +
        stacktrace[trace_depth_start].functionName.underline  + ": "+ type_data + "\n" +
		message + "\n" +
		exiting_process_or_thread_string_below + stacktrace[trace_depth_start].fileName + ":" + stacktrace[trace_depth_start].lineNumber; 

        
    }
    else
    {
        log_info = "\n[Exiting "+process_or_thread+"]-> ".red + ": " + message + "\n";
    }

    // display the message
    console.log(log_info);
	
	// sleep a sec if exiting from a worker thread
	if(!isMainThread)
	{
		await asyncSleepMiliseconds(50);
	}
    
    // exit the process
    process.exit(0);

}

// export exit
exported.exit = exit;

// return the exported object containing all utility functions
return exported;

};